#include "Initiallization.h"


#include <cstring>
#include <cstdlib>
#include <Eigen/Dense>

using namespace Eigen;
using namespace std;



void Set_BS_Location();
void Initialize_CHANNEL();
double Get_distance(LOCATION, LOCATION);


void CQI_Map_creation_Map_connection();
void Schedule_Map_creation_Connection();
void Codebook_Generation(void);
void ClearMAPs2StartADrop();
void Link_configuration();
void Get_CouplingLoss();

void Declare_Precalc_Parameter();


void Do_precalculation();
void Load_precalculation_file();
void Print_DAT_FILE();
void Do_FT_precalc();
void Do_PMI_precalc(int );
void Do_SINR_estimate_precalc(int);
void Do_CH_COEF_SUM_precalc(int);

void Load_CHIR_precalc_file();
//void Load_FT_precalc_file();
void Load_PMI_precalc_file();
void Load_SINR_estimate_precalc_file();
void Load_CH_COEF_SUM_precalc_file();
//void Load_BLER_vienna_file();


void Set_Parameter(int scenario)
{

	/*=====================================================================================
	INPUT: scenario index (1=InH, 2=UMi, 3=UMa, 4=RMa, 5=SMa, 6=3D-UMi, 7=3D-UMa, 8=3D-UMa-H)
	OUTPUT: scenario parameter(num_BS, num_MS, Freq, bandwidth, inter_site_distance.)
	5 ó  ϳ ϰ, ׿  parameter Ѵ.
	====================================================================================*/

	TYPE = scenario;
	
	if (TYPE == 11) // InH
	{
		if (Configuration_Type == 0) ///// InH config A
		{
			///// IMT-2020 EVal, Page 19

			num_LINK = num_MS;

			//W = 20;//street width
			//h = 20;//avg. building height

			user_speed = 3;  //[3km/h]
			Total_BS_Tx_power = 21; //[dBm for 20MHz]
			Frequency = 4000000000; //[4GHz] 4 * 10^9
			Wavelength = light_speed / Frequency;
			bs_maxpower = 21; //[dBm for 20MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 5;
			MS_noisefig = 7;
			ue_antenna_element_gain = 0;
			max_antgain = 5; //[dBi]  BS antenna element gain

			bandwidth = 10000000; //[10MHz]  10 * 10^6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 3.; //[m]
			ms_height = 1.5;
			inter_site_distance = 20;
			min_distance = 0.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 1;
			num_propagation_condition = 2; // LOS / NLOS

			///////////////////////////////////////////////////
		}
		else if (Configuration_Type == 1) ///// InH config B
		{
			///// IMT-2020 EVal, Page 19

			num_LINK = num_MS;

			//W = 20;//street width
			//h = 20;//avg. building height

			user_speed = 3;  //[3km/h]  
			Total_BS_Tx_power = 20; //[dBm for 40MHz]
			Frequency = 30000000000; //[30GHz]
			Wavelength = light_speed / Frequency;
			bs_maxpower = 20; //[dBm for 40MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 7;
			MS_noisefig = 10;
			ue_antenna_element_gain = 5;
			max_antgain = 5; //[dBi]  BS antenna element gain

			bandwidth = 40000000; //[40MHz]  40 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 3.; //[m]
			ms_height = 1.5;
			inter_site_distance = 20;
			min_distance = 0.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 1;
			num_propagation_condition = 2; // LOS / NLOS

			///////////////////////////////////////////////////
		}
		else if (Configuration_Type == 2) ///// InH config C
		{
			///// IMT-2020 EVal, Page 19

			num_LINK = num_MS;

			//W = 20;//street width
			//h = 20;//avg. building height

			user_speed = 3;  //[3km/h]  
			Total_BS_Tx_power = 18; //[dBm for 40MHz]
			Frequency = 70000000000; //[70GHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 18; //[dBm for 40MHz] =Total BS TX power
			ms_maxpower = 21; //[dBm UE power class]
			BS_noisefig = 7;
			MS_noisefig = 10;
			ue_antenna_element_gain = 5;
			max_antgain = 5; //[dBi]  BS antenna element gain

			bandwidth = 40000000; //[40MHz]  40 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)

			bs_height = 3.; //[m]
			ms_height = 1.5;

			inter_site_distance = 20;

			min_distance = 0.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 1;
			num_propagation_condition = 2; // LOS / NLOS

			///////////////////////////////////////////////////
		}
		else
		{
			cout << "Configuration Type Error !!" << endl;
			getchar();
		}

	}
	else if (TYPE == 12) // Dense_Urban
	{
		if (Configuration_Type == 0) ///// Dense Urban config A
		{
			///// IMT-2020 EVal, Page 20

			num_LINK = num_MS;

			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 41; //[dBm for 10MHz]
			Frequency = 4000000000; //[4GHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 41; //[dBm for 10MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 5;
			MS_noisefig = 7;
			ue_antenna_element_gain = 0;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 10000000; //[10MHz]  10 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)

			bs_height = 25.; //[m]
			ms_height_in = 0;   /// in LOS prob part
			ms_height_out = 1.5;
			inter_site_distance = 200;

			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 8;
			num_propagation_condition = 3;     /// LOS, NLOS, OUT2IN

			//W = 20.; //street width
			//h = 20.; //avg. building height


			//ANGLE_tilt = 6.;

			////////////////////////////////////////////////

			//MAX_v_angle = atan2((bs_height - ms_height_out), min_distance) * 180 / pi;   // FOR Pre-Calculate v angle


		}
		else if (Configuration_Type == 1) ///// Dense Urban config B
		{
			///// IMT-2020 EVal, Page 20

			num_LINK = num_MS;

			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 37; //[dBm for 40MHz]
			Frequency = 30000000000; //[30GHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 37; //[dBm for 40MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 7;
			MS_noisefig = 10;
			ue_antenna_element_gain = 5;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 40000000; //[40MHz]  40 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 25.; //[m]

			ms_height_in = 0;   /// in LOS prob part
			ms_height_out = 1.5;
			inter_site_distance = 200;
			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 8;
			num_propagation_condition = 3;

		}
		else if (Configuration_Type == 2) ///// Dense Urban config C Macro - 30GHz
		{
			///// IMT-2020 EVal, Page 20
			///// Macro  cell case
			num_LINK = num_MS;

			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 37; //[dBm for 40MHz]
			Frequency = 30000000000; //[30GHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 37; //[dBm for 40MHz]=Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 7;
			MS_noisefig = 10;
			ue_antenna_element_gain = 5;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 40000000; //[40MHz]  40 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 25.; //[m]

			ms_height_in = 0;   /// in LOS prob part
			ms_height_out = 1.5;
			inter_site_distance = 200;
			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 8;
			num_propagation_condition = 3;

		}
		else if (Configuration_Type == 3) ///// Dense Urban config C Micro - 4GHz
		{
			///// IMT-2020 EVal, Page 20
			///// Micro  cell case
			num_LINK = num_MS;

			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 30; //[dBm for 10MHz]
			Frequency = 4000000000; //[4GHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 30; //[dBm for 10MHz] = Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 5;
			MS_noisefig = 7;
			ue_antenna_element_gain = 0;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 10000000; //[10MHz]  10 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)

			bs_height = 10.; //[m]
			ms_height_in = 0;   /// in LOS prob part
			ms_height_out = 1.5;
			inter_site_distance = 100;   /// should be reported by the proponent... 10 = just random number

			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 8;
			num_propagation_condition = 3;     /// LOS, NLOS, OUT2IN

		}
		else
		{
			cout << "Configuration Type Error !!" << endl;
			getchar();
		}

	}
	else if (TYPE == 13) // RMa
	{
		if (Configuration_Type == 0) ///// RMa config A
		{
			///// IMT-2020 EVal, Page 21

			num_LINK = num_MS;

			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 46; //[dBm for 10MHz]
			Frequency = 700000000; //[700MHz] 700 * 10^6
			Wavelength = light_speed / Frequency;

			bs_maxpower = 46; //[dBm for 10MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 5;
			MS_noisefig = 7;
			ue_antenna_element_gain = 0;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 10000000; //[10MHz]  10 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 35.; //[m]

			ms_height_out = 1.5;
			ms_height_in = 1.5;

			inter_site_distance = 1732;

			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 1;
			num_propagation_condition = 3;

		}
		else if (Configuration_Type == 1) ///// RMa config B
		{
			///// IMT-2020 EVal, Page 21

			num_LINK = num_MS;

			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 46; //[dBm for 10MHz]
			Frequency = 4000000000; //[4GHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 46; //[dBm for 10MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 5;
			MS_noisefig = 7;
			ue_antenna_element_gain = 0;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 10000000; //[10MHz]  10 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 35.; //[m]

			ms_height_out = 1.5;
			ms_height_in = 1.5;

			inter_site_distance = 1732;

			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 1;
			num_propagation_condition = 3;

		}
		else if (Configuration_Type == 2) ///// RMa config C
		{
			///// IMT-2020 EVal, Page 21

			num_LINK = num_MS;


			user_speed = 3;  //[3km/h] Indoor user // outdoor user speed -> decided later
			Total_BS_Tx_power = 46; //[dBm for 10MHz]
			Frequency = 700000000; //[700MHz]
			Wavelength = light_speed / Frequency;

			bs_maxpower = 46; //[dBm for 10MHz] =Total BS TX power
			ms_maxpower = 23; //[dBm UE power class]
			BS_noisefig = 5;
			MS_noisefig = 7;
			ue_antenna_element_gain = 0;
			max_antgain = 8; //[dBi]  BS antenna element gain

			bandwidth = 10000000; //[10MHz]   10 * 10 ^ 6
			noise = dBm2linear(thermal_noise) * 15000; //[linear]  // 15000 = UE bandwidth (15kHz)
			bs_height = 35.; //[m]

			ms_height_out = 1.5;
			ms_height_in = 1.5;

			inter_site_distance = 6000;

			min_distance = 10.; // not decided in M.2412
			ANGLE_tilt = 0.; // change by calibration config

			num_floor = 1;
			num_propagation_condition = 3;

		}
		else
		{
			cout << "Configuration Type Error !!" << endl;
			getchar();
		}

	}


	cout << "carrier_freq              : " << Frequency << endl;
	cout << "inter_site_distance       : " << inter_site_distance << endl;
	cout << "secter_to_ue_min_distance : " << min_distance << endl;
	cout << "indoor_ue_speed           : " << user_speed << endl;
	cout << "h_BS                      : " << bs_height << endl;
	cout << "bs_noise_figure           : " << BS_noisefig << endl;
	cout << "ue_noise_figure           : " << MS_noisefig << endl;
	cout << "bs_antenna_element_gain   : " << max_antgain << endl;
	cout << "ue_antenna_element_gain   : " << ue_antenna_element_gain << endl;
	cout << "thermal_nosie_level       : " << thermal_noise << endl;
	cout << "simulation_bandwidth      : " << bandwidth << endl;

	getchar();


	bs = new BS[num_BS];
	ms = new MS[num_MS];
	link = new LINK[num_LINK];
	sector = new Sector[num_SECTORS];
	channel = new CHANNEL*[num_BS];

	for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
	{
		channel[bs_idx] = new CHANNEL[num_MS];
	}

	Set_BS_Location();
	

}
void Set_BS_Location()
{
	/*
	INPUT: scenario index 
	OUTPUT: BS location
	ó  BS ġ 
	*/


	if (TYPE == 11) //// 5G InH
	{

		bs[0].loc.x = -50;
		bs[0].loc.y = 35;

		bs[1].loc.x = -30;   
		bs[1].loc.y = 35;

		bs[2].loc.x = -10;
		bs[2].loc.y = 35;

		bs[3].loc.x = 10;
		bs[3].loc.y = 35;

		bs[4].loc.x = 30;
		bs[4].loc.y = 35;

		bs[5].loc.x = 50;
		bs[5].loc.y = 35;

		bs[6].loc.x = -50;
		bs[6].loc.y = 15;

		bs[7].loc.x = -30;
		bs[7].loc.y = 15;

		bs[8].loc.x = -10;
		bs[8].loc.y = 15;

		bs[9].loc.x = 10;
		bs[9].loc.y = 15;

		bs[10].loc.x = 30;
		bs[10].loc.y = 15;

		bs[11].loc.x = 50;
		bs[11].loc.y = 15;

	}
	else  //// hexagonal scenario
	{

		double cell_radius = inter_site_distance * (1 / sqrt(3));

		bs[0].loc.x = 0;
		bs[1].loc.x = 0;
		bs[2].loc.x = -1.5 * cell_radius;
		bs[3].loc.x = -1.5 * cell_radius;
		bs[4].loc.x = 0;
		bs[5].loc.x = 1.5 * cell_radius;
		bs[6].loc.x = 1.5 * cell_radius;
		bs[7].loc.x = 0;
		bs[8].loc.x = -1.5 * cell_radius;
		bs[9].loc.x = -3. * cell_radius;
		bs[10].loc.x = -3. * cell_radius;
		bs[11].loc.x = -3. * cell_radius;
		bs[12].loc.x = -1.5 * cell_radius;
		bs[13].loc.x = 0;
		bs[14].loc.x = 1.5 * cell_radius;
		bs[15].loc.x = 3. * cell_radius;
		bs[16].loc.x = 3. * cell_radius;
		bs[17].loc.x = 3. * cell_radius;
		bs[18].loc.x = 1.5 * cell_radius;


		//Location y

		bs[0].loc.y = 0;
		bs[1].loc.y = inter_site_distance;
		bs[2].loc.y = inter_site_distance * (0.5);
		bs[3].loc.y = inter_site_distance * (-0.5);
		bs[4].loc.y = inter_site_distance * (-1);
		bs[5].loc.y = inter_site_distance * (-0.5);
		bs[6].loc.y = inter_site_distance * (0.5);
		bs[7].loc.y = inter_site_distance * (2.);
		bs[8].loc.y = inter_site_distance * (1.5);
		bs[9].loc.y = inter_site_distance * (1.);
		bs[10].loc.y = 0;
		bs[11].loc.y = inter_site_distance * (-1);
		bs[12].loc.y = inter_site_distance * (-1.5);
		bs[13].loc.y = inter_site_distance * (-2);
		bs[14].loc.y = inter_site_distance * (-1.5);
		bs[15].loc.y = inter_site_distance * (-1);
		bs[16].loc.y = 0;
		bs[17].loc.y = inter_site_distance * (1.);
		bs[18].loc.y = inter_site_distance * (1.5);
	}



}

void Initialdrop()
{

	///double cell_radius = inter_site_distance / sqrt(3.);

	///////////////////////////////////////////////////////////////// 5G InH
	if (TYPE == 11)          
	{
		for (int bs_idx = 0; bs_idx < num_BS; bs_idx++) /// 12
		{
		
			for (int ms_idx = 0; ms_idx < num_MS_persector; ms_idx++)  /// 10
			{
				ms[ms_idx + bs_idx * num_MS_persector].nearest_bs_idx = bs_idx;   // ms 0~9 -> bs 0, ms 10~19 -> bs 1

				do
				{
					ms[ms_idx + bs_idx * num_MS_persector].loc.x = (bs[bs_idx].loc.x - 10) + 20. * randnum.u();

					if (bs_idx < 6)   /// 6 BS  on Top line  (0 ~ 5)
					{
						ms[ms_idx + bs_idx * num_MS_persector].loc.y = (bs[bs_idx].loc.y - 10) + 25. * randnum.u();
					}
					else  /// 6 BS on Bottom line (6 ~ 11)
					{
						ms[ms_idx + bs_idx * num_MS_persector].loc.y = (bs[bs_idx].loc.y - 15) + 25. * randnum.u();
					}

					ms[ms_idx + bs_idx * num_MS_persector].moving_direction = 2 * 180 * randnum.u() - 180;
					ms[ms_idx + bs_idx * num_MS_persector].speed = user_speed * 1000 / 3600;
					ms[ms_idx + bs_idx * num_MS_persector].Indoor = true;

				} while (Get_distance(ms[ms_idx + bs_idx * num_MS_persector].loc, bs[bs_idx].loc) < min_distance);

			}
		}
	}
	////////////////////////////////////////////////////////////////// hexagonal scenario
	else  
	{
		for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
		{
			for (int sector_idx = 1; sector_idx <= 3; sector_idx++)
			{
				for (int ms_idx = bs_idx*num_MS_persector * 3 + num_MS_persector*(sector_idx - 1); ms_idx < bs_idx*num_MS_persector * 3 + num_MS_persector*sector_idx; ms_idx++) // ms_idx = 0~9 -> bs[0],sector1, 10~19 -> bs[0],sector2, 20~29 -> bs[0],sector3, 30~39 -> bs[1],sector1 ......
				{
					ms[ms_idx].Configuration(ms_idx, sector_idx);
					ms[ms_idx].nearest_bs_idx = bs_idx;
					ms[ms_idx].loc.x = bs[bs_idx].loc.x + ms[ms_idx].loc.x;
					ms[ms_idx].loc.y = bs[bs_idx].loc.y + ms[ms_idx].loc.y;
					ms[ms_idx].moving_direction = 2 * 180 * randnum.u() - 180; //travel angle (elevation)
					ms[ms_idx].moving_direction_azimuth = 2 * 180 * randnum.u() - 180; // travel azimuth angle
					ms[ms_idx].speed = user_speed * 1000 / 3600;

					if (TYPE == 12)   /// imt 2020 Dense urban
					{
						if (ms[ms_idx].Indoor == false)  //// outdoor speed
						{
							ms[ms_idx].speed = 30. * 1000. / 3600.;
						}

					}
					else if (TYPE == 13)  //// imt 2020 rural
					{
						if (ms[ms_idx].Indoor == false)  /// outdoor speed
						{
							if (Configuration_Type == 2) /// configuration C
							{
								if (randnum.u() < 0.5) 
								{
									ms[ms_idx].speed = user_speed * 1000. / 3600.;  /// outdoor pedestrian 3km/h
								}
								else
								{
									ms[ms_idx].speed = 30. * 1000. / 3600.;  /// outdoor in car
								}
							}
							else
							{
								ms[ms_idx].speed = 120. * 1000. / 3600.;
							}
						}
					}
				}
			}
		}
	}
}

void Initialize_CHANNEL()
{
	double cell_radius = inter_site_distance / sqrt(3.);
	double distance;

	LOCATION temp_bs;
	LOCATION nearest_bs;
	
	//////////////////////////////////////////////////////////////////////////////// 5G InH
	if (TYPE == 11) 
	{
		for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
		{
			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
			{
				nearest_bs.x = bs[bs_idx].loc.x;
				nearest_bs.y = bs[bs_idx].loc.y;

				distance = Get_distance(ms[ms_idx].loc, nearest_bs);

				channel[bs_idx][ms_idx].self_bs.x = nearest_bs.x;
				channel[bs_idx][ms_idx].self_bs.y = nearest_bs.y;
				channel[bs_idx][ms_idx].self_ms.x = ms[ms_idx].loc.x;
				channel[bs_idx][ms_idx].self_ms.y = ms[ms_idx].loc.y;

				channel[bs_idx][ms_idx].Set_channel(distance, ms[ms_idx].Indoor, bs_idx, ms_idx, channel[bs_idx][ms_idx].self_bs, channel[bs_idx][ms_idx].self_ms);
			}
		}
	}
	////////////////////////////////////////////////////////////////////////////////// 2 tier wrap around Hexagonal grid , 3 Sector per 1 BS
	else 
	{
		for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
		{
			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)  // Ÿ ؼ  ms wrap around߿   bs 19 ã
			{
				nearest_bs.x = bs[bs_idx].loc.x;
				nearest_bs.y = bs[bs_idx].loc.y;

				temp_bs.x = bs[bs_idx].loc.x + 4.5*cell_radius;                   //C 1
				temp_bs.y = bs[bs_idx].loc.y + 7. * (sqrt(3.) / 2.) * cell_radius;
				if (Get_distance(ms[ms_idx].loc, nearest_bs) > Get_distance(ms[ms_idx].loc, temp_bs))
				{
					nearest_bs.x = temp_bs.x;
					nearest_bs.y = temp_bs.y;
				}

				temp_bs.x = bs[bs_idx].loc.x - 3.*cell_radius;                  // c 2
				temp_bs.y = bs[bs_idx].loc.y + 8. * (sqrt(3.) / 2.) * cell_radius;
				if (Get_distance(ms[ms_idx].loc, nearest_bs) > Get_distance(ms[ms_idx].loc, temp_bs))
				{
					nearest_bs.x = temp_bs.x;
					nearest_bs.y = temp_bs.y;
				}

				temp_bs.x = bs[bs_idx].loc.x - 7.5*cell_radius;                // c 3
				temp_bs.y = bs[bs_idx].loc.y + (sqrt(3.) / 2.) * cell_radius;
				if (Get_distance(ms[ms_idx].loc, nearest_bs) > Get_distance(ms[ms_idx].loc, temp_bs))
				{
					nearest_bs.x = temp_bs.x;
					nearest_bs.y = temp_bs.y;
				}

				temp_bs.x = bs[bs_idx].loc.x - 4.5*cell_radius;               // c 4
				temp_bs.y = bs[bs_idx].loc.y - 7. * (sqrt(3.) / 2.) * cell_radius;
				if (Get_distance(ms[ms_idx].loc, nearest_bs) > Get_distance(ms[ms_idx].loc, temp_bs))
				{
					nearest_bs.x = temp_bs.x;
					nearest_bs.y = temp_bs.y;
				}

				temp_bs.x = bs[bs_idx].loc.x + 3.*cell_radius;                  // c 5
				temp_bs.y = bs[bs_idx].loc.y - 8. * (sqrt(3.) / 2.) * cell_radius;
				if (Get_distance(ms[ms_idx].loc, nearest_bs) > Get_distance(ms[ms_idx].loc, temp_bs))
				{
					nearest_bs.x = temp_bs.x;
					nearest_bs.y = temp_bs.y;
				}

				temp_bs.x = bs[bs_idx].loc.x + 7.5*cell_radius;                 //c 6 
				temp_bs.y = bs[bs_idx].loc.y - (sqrt(3.) / 2.) * cell_radius;
				if (Get_distance(ms[ms_idx].loc, nearest_bs) > Get_distance(ms[ms_idx].loc, temp_bs))
				{
					nearest_bs.x = temp_bs.x;
					nearest_bs.y = temp_bs.y;
				}

				distance = Get_distance(ms[ms_idx].loc, nearest_bs); // ߿   bs Ÿ 


				channel[bs_idx][ms_idx].self_bs.x = nearest_bs.x;
				channel[bs_idx][ms_idx].self_bs.y = nearest_bs.y;
				channel[bs_idx][ms_idx].self_ms.x = ms[ms_idx].loc.x;
				channel[bs_idx][ms_idx].self_ms.y = ms[ms_idx].loc.y;

				channel[bs_idx][ms_idx].Set_channel(distance, ms[ms_idx].Indoor, bs_idx, ms_idx, channel[bs_idx][ms_idx].self_bs, channel[bs_idx][ms_idx].self_ms);
			}
		}
	}
}

void Link_configuration()
{
	////////////////////////////////////////////////////////////////////////////////////////////////////
	for (int idx = 0; idx < num_SECTORS; idx++)
	{
		sector[idx].ue_in_control.clear();
		sector[idx].W = NULL;
		sector[idx].PMI_read = NULL;

	}

	///////////////////////////////////////////////////////////////////////////////////////////////////
	for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
	{
		cout << "link_config ms_idx = " << ms_idx + 1 << "/570" << endl;
		link[ms_idx].Configuration(ms_idx);

		sector[link[ms_idx]._sector_in_control].ue_in_control.push_back(ms_idx);  // Ͱ ϴ ms_idx ue_in_control ִ´ //   ġ   10 ־, ׷ ϸ Ʋ. 
	}
}

void InitializeSystem()
{
	CQI_Map_creation_Map_connection();
	/*
	INPUT: 
	OUTPUT: PMI_map, CQI_map
	Notes: PMI_map, CQI_map ϰ, MS class PMI,CQI map 
	*/
	Schedule_Map_creation_Connection();
	/*
	INPUT: 
	OUTPUT: Schedule_map
	Notes: Schedule_map ϰ, Sector class schedulewrite 
	*/
	Codebook_Generation();
	/*
	INPUT: 
	OUTPUT: code_word
	Notes: code book generation
	*/

}




/*===================================================================
FUNCTION: CQI_Map_creation_Map_connection()
===================================================================*/
void CQI_Map_creation_Map_connection()
{
	// CQI Map creation
	pppPMI_map = new int        **[num_MS];
	pppCQI_Map = new double     **[num_MS];

	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{
		pppPMI_map[ue_idx] = new int        *[num_rbs_in_1ms];
		pppCQI_Map[ue_idx] = new double     *[num_rbs_in_1ms];
	}

	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			pppPMI_map[ue_idx][freq_idx] = new int[cqi_history_length];
			pppCQI_Map[ue_idx][freq_idx] = new double[cqi_history_length];
		}
	}

	// CQI Map to MS connection

	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{
		ms[ue_idx].PMI = pppPMI_map[ue_idx];
		ms[ue_idx].CQI = pppCQI_Map[ue_idx];
	}


}


/*===================================================================
FUNCTION: Schedule_Map_creation_Connection()
===================================================================*/
void Schedule_Map_creation_Connection()
{

	// Schedule Map Creation

	pppSchedule_Map = new SCHEDULE_DECISION **[num_SECTORS];

	for (int sector_idx = 0; sector_idx < num_SECTORS; sector_idx++)
	{
		pppSchedule_Map[sector_idx] = new SCHEDULE_DECISION *[num_rbs_in_1ms];

		for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
		{
			pppSchedule_Map[sector_idx][rb_idx] = new SCHEDULE_DECISION[NUM_TX];
		}
	}


	// Schedule Map to Sector Connection

	for (int sector_idx = 0; sector_idx < num_SECTORS; sector_idx++)
	{
		sector[sector_idx].ppschedulewrite = pppSchedule_Map[sector_idx];
	}




}

void Codebook_Generation(void) 
{
	// MatrixXcd =  (dynamic,dynamic) complex double matrix
	// Matrix2cd =  (2,2)             complex double matrix

	//typedef Matrix<complex< double>, 1, 4 > u[16];
	/*
	u[0] = "1 -1 -1 -1";
	u[1] = "1 0-1i 1 0+1i";
	u[2] = "1 1 -1 1 ";
	u[3] = "1 0-1i 1 1-1i";
	u[4] = "1 -0.707-0.707i 0-1i 0.707-0.707i";
	u[5] = "1 0.707-0.707i 0+1i -0.707-0.707i";
	u[6] = "1 0.707+0.707i 1-1i 0.707+0.707i";
	u[7] = "1 -0.707+0.707i 0+1i 0.707+0.707i";
	u[8] = "1 -1 1 1";
	u[9] = "1 1-1i -1 1-1i";
	u[10] = "1 1 1 -1";
	u[11] = "1 1+1i -1 1+1i";
	u[12] = "1 -1 -1 1";
	u[13] = "1 -1 1 -1";
	u[14] = "1 1 -1 -1";
	u[15] = "1 1 1 1";
	*/

	RowVector4cd u[16];
	Vector4cd u_T[16];
	Matrix4cd I = Matrix4cd::Identity();


	//header 
	//Matrix4cd WW[16];
	//Vector4cd code_word[16];


	// 4 Antenna Codebook

	u[0] << complex<double>(1, 0), complex<double>(-1, 0), complex<double>(-1, 0), complex<double>(-1, 0);
	u[1] << complex <double>(1, 0), complex <double>(0, -1), complex <double>(1, 0), complex <double>(0, 1);
	u[2] << complex <double>(1, 0), complex <double>(1, 0), complex <double>(-1, 0), complex <double>(1, 0);
	u[3] << complex <double>(1, 0), complex <double>(0, -1), complex <double>(1, 0), complex <double>(1, -1);
	u[4] << complex <double>(1, 0), complex <double>(-0.707, -0.707), complex <double>(0, -1), complex <double>(0.707, -0.707);
	u[5] << complex <double>(1, 0), complex <double>(0.707, -0.707), complex <double>(0, 1), complex <double>(-0.707, -0.707);
	u[6] << complex <double>(1, 0), complex <double>(0.707, 0.707), complex <double>(1, -1), complex <double>(0.707, 0.707);
	u[7] << complex <double>(1, 0), complex <double>(-0.707, 0.707), complex <double>(0, 1), complex <double>(0.707, 0.707);
	u[8] << complex <double>(1, 0), complex <double>(-1, 0), complex <double>(1, 0), complex <double>(1, 0);
	u[9] << complex <double>(1, 0), complex <double>(1, -1), complex <double>(-1, 0), complex <double>(1, -1);
	u[10] << complex <double>(1, 0), complex <double>(1, 0), complex <double>(1, 0), complex <double>(-1, 0);
	u[11] << complex <double>(1, 0), complex <double>(1, 1), complex <double>(-1, 0), complex <double>(1, 1);
	u[12] << complex <double>(1, 0), complex <double>(-1, 0), complex <double>(-1, 0), complex <double>(1, 0);
	u[13] << complex <double>(1, 0), complex <double>(-1, 0), complex <double>(1, 0), complex <double>(-1, 0);
	u[14] << complex <double>(1, 0), complex <double>(1, 0), complex <double>(-1, 0), complex <double>(-1, 0);
	u[15] << complex <double>(1, 0), complex <double>(1, 0), complex <double>(1, 0), complex <double>(1, 0);

	for (int i = 0; i < 16; i++)
	{
		u_T[i] = u[i].transpose();

	}

	// Get Precoding Matrix WW
	///Eigen Library adjoint() = hermitian()
	for (int i = 0; i < 16; i++)
	{
		WW[i] = I - 2 * (u_T[i] * u_T[i].adjoint()) / ((u_T[i].adjoint() * u_T[i])(0, 0));


		//int column_number = (int)(4 * randnum.u());



		//code_word[i] = WW[i].col(0); // codeword precoding matrix ù Į ߴ
		//code_word[i] = WW[i].col(column_number); // codeword precoding matrix Į  ϰ ϴ° غ

        // codeword 64  (ڵ PMI  )
		code_word[4 * i] = WW[i].col(0);
		code_word[4 * i + 1] = WW[i].col(1);
		code_word[4 * i + 2] = WW[i].col(2);
		code_word[4 * i + 3] = WW[i].col(3);

	}


	/*
	cout << "WW[0]" << endl << WW[0] << endl;
	cout << "WW[1]" << endl << WW[1] << endl;
	cout << "WW[4]" << endl << WW[4] << endl;
	cout << "WW[15]" << endl << WW[15] << endl;
	*/

}

void ClearMAPs2StartADrop()
{
	/*
	INPUT: 
	OUTPUT: 
	Notes: CQi_map Schedule_map ʱȭ, CQImap 0 Schedulemap -1 (no_decision) ʱȭ
	*/

	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			for (int time_idx = 0; time_idx < cqi_history_length; time_idx++)
			{
				pppCQI_Map[ue_idx][freq_idx][time_idx] = 0;
			}
		}
	}

	SCHEDULE_DECISION no_decision;

	no_decision.ue_selected = NO_UE;
	no_decision.mcs_selected = NO_CQI;
	no_decision.cqi_selected = NO_CQI;
	no_decision.capacity = -1;

	for (int sector_idx = 0; sector_idx < num_SECTORS; sector_idx++)
	{
		for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
		{
			for (int stream = 0; stream < NUM_TX; stream++)
			{
				pppSchedule_Map[sector_idx][freq_idx][stream] = no_decision;
			}
		}
	}

	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{
		ms[ue_idx].ppSchedulerRead = pppSchedule_Map[link[ue_idx]._sector_in_control];
	}	
}

void Get_CouplingLoss()
{
	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{

		double linear_signal = dB2linear(link[ue_idx].str_signal);
		double linear_interference = dB2linear(link[ue_idx].interference);

		linear_interference = linear_interference + noise;

		double geometry = linear2dB(linear_signal / linear_interference);

		Get_Geometry(geometry, ue_idx);

		//double cp = max_antgain - link[ue_idx].incar_loss - BS_feeder_loss + link[ue_idx].static_gain[link[ue_idx]._sector_in_control]; // static gain = antgain -(pathloss + shadowing fading)

		if (TYPE == 11) // indoor case
		{
			double cp = link[ue_idx].str_signal - bs_maxpower;  // str_signal = bs_maxpower + max_antgain + combined antgain - pathloss
			//double cp = -link[ue_idx].link_pathloss - link[ue_idx].link_antgain;  // str_signal = bs_maxpower + max_antgain + combined antgain - pathloss

			Coupling_Loss_Get(cp, ue_idx);
		}
		else
		{
			//double cp = -link[ue_idx].incar_loss - link[ue_idx].Otoi_loss + link[ue_idx].str_signal - bs_maxpower;  // str_signal = bs_maxpower + max_antgain - combined antgain - pathloss


			double cp = link[ue_idx].str_signal - bs_maxpower;  // str_signal = bs_maxpower + max_antgain - combined antgain - pathloss

			//double cp =  -1 * link[ue_idx].link_coupling_loss;  // str_signal = bs_maxpower + max_antgain - combined antgain - pathloss

			Coupling_Loss_Get(cp, ue_idx);

		}
		/*
		double rms = link[ue_idx].link_RMS_delay_spread * pow(10.,6.); // micro second
		Get_RMS_delay_spread(rms, ue_idx);
		double aoa_spread = link[ue_idx].link_AOA_spread;// *(pi / 180.);
		Get_circular_angle_spread_AOA(aoa_spread, ue_idx);
		double aod_spread = link[ue_idx].link_AOD_spread;// *(pi / 180.);
		Get_circular_angle_spread_AOD(aod_spread, ue_idx);
		*/
	}
}




double Get_distance(LOCATION a, LOCATION b)
{
	return sqrt((a.x - b.x)*(a.x - b.x) + (a.y - b.y)*(a.y - b.y));
}




/*===================================================================
FUNCTION: loading()

Print loading percent
===================================================================*/
void loading(void)
{
	
	//system("clear");
	//system("cls");
	
	if (t < SCHEDULE_DELAY)
	{
		cout << " Initial Setting is going on " << abs(t - SCHEDULE_DELAY) << endl;
	}
	else
	{
		cout << " Pleas Wait Until it get to the 100% " << endl;
		cout << "------------------------" << "\t"
			<< 100.*(double)(t - SCHEDULE_DELAY + 1 + drop_idx*(N_pf + run_times)) / (double)((N_pf + run_times)*num_drop) << "%" << "\t"
			<< "------------------------" << endl;
		cout << " Current Drop is " << drop_idx + 1 << "/" << num_drop << endl;
		cout << "------------------------" << "\t"
			<< 100.*(double)(t - SCHEDULE_DELAY + 1) / (double)(N_pf + run_times) << "%" << "\t"
			<< "------------------------" << endl;
		cout << " Output File name is `" << file_name << "`" << endl;
	}
	

}

void Killing_Memory(void)
{
	/*
	for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
	{
		ms[ue_idx].Kill_Memory();
	}
	*/


	for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
	{
		for (int ue_idx = 0; ue_idx < num_MS; ue_idx++)
		{
			channel[bs_idx][ue_idx].Kill_Memory();
		}
		
	}

	

	if (CH_CAL == 0 || CH_CAL == 3)  //// 0 = Pre-calculate & run , 3 = ܺ  Է¹
	{

		//for (int time = 0; time < precalculate_time; time++)        // ð 20ʸŭ
		//{
			for (int los = 0; los < 2; los++)            // los 
			{
				for (int indoor = 0; indoor < 2; indoor++)   /////// indoor 
				{
					for (int v = 0; v < anglev_mapping_value; v++)
					{
						for (int u = 0; u < NUM_TX; u++)         // TX
						{
							for (int s = 0; s < NUM_RX; s++) ////// RX 
							{
								delete[] CHIR_precalc[los][indoor][v][u][s];
							}

							delete[] CHIR_precalc[los][indoor][v][u];
						}

						delete[] CHIR_precalc[los][indoor][v];
					}
					delete[] CHIR_precalc[los][indoor];
				}
			}
		//}

		//for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
		//{
			for (int los_idx = 0; los_idx < 2; los_idx++)
			{
				for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
				{
					for (int v = 0; v < anglev_mapping_value; v++)
					{
						delete[] Hm[los_idx][indoor_idx][v];

					}
					delete[] Hm[los_idx][indoor_idx];
				}
			}
		//}


		for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
		{
			for (int los_idx = 0; los_idx < 2; los_idx++)
			{
				for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
				{
					for (int v = 0; v < anglev_mapping_value; v++)
					{
						//for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
						//{
							//delete[] PMI_precalc[time_idx][los_idx][indoor_idx][v][freq_idx];
						//}

						delete[] PMI_precalc[time_idx][los_idx][indoor_idx][v];
					}
					delete[] PMI_precalc[time_idx][los_idx][indoor_idx];
				}
			}
		}

		/////////////////////////
		for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
		{
			for (int los_idx = 0; los_idx < 2; los_idx++)
			{
				for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
				{
					for (int v = 0; v < anglev_mapping_value; v++)
					{
						//for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
						//{
						//	delete[] sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v][freq_idx];
						//}

						delete[] sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v];
					}
					delete[] sinr_estimate_precalc[time_idx][los_idx][indoor_idx];
				}
			}
		}
		///////////////////////
		//////////////////////////////////
		for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
		{
			for (int los_idx = 0; los_idx < 2; los_idx++)
			{
				for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
				{
					for (int v = 0; v < anglev_mapping_value; v++)
					{
						delete[] sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v];
					}
					delete[] sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx];
				}
			}
		}


	}

	



	/*
	for (int sec_idx = 0; sec_idx < num_SECTORS; sec_idx++)
	{
		//sector[sec_idx].Kill_Memory();
	}
		*/
}



void Parameter_initialization() 
{
	//// initialization per drop
	for (int idx = 0; idx < num_MS; idx++)
	{
		ms[idx].H_m = NULL;
		ms[idx].HI_m = NULL;
		ms[idx].downlinkaveragedThroghput = 0;
		ms[idx].downlinkHARQeSINR = 0;
		ms[idx].downlinkThroghput = 0;

		/*
		ms[idx].failed = 0;
		ms[idx].total_num_tx = 0;
		ms[idx].BLER_SUM_Value = 0;
		ms[idx].BLER_SUM_count = 0;
		ms[idx].decision_BLER_SUM_Value = 0;
		ms[idx].decision_BLER_SUM_count = 0;
		*/

	}

	if (drop_idx == 0)  /// Declare only first drop
	{
		//////////////////////// Declare CHIR /////////////////
		///////////////////////////////////////////////////////
		if (TYPE == 11)
		{
			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++) // 
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					channel[bs_idx][ms_idx].CHIR = new complex <double>**[NUM_TX];

					for (int u = 0; u < NUM_TX; u++)
					{
						channel[bs_idx][ms_idx].CHIR[u] = new complex <double>*[NUM_RX];

						for (int s = 0; s < NUM_RX; s++)
						{
							channel[bs_idx][ms_idx].CHIR[u][s] = new complex <double>[MAX_NUM_CLUSTERS];

							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}

			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++) // 
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					channel[bs_idx][ms_idx].CHIR_LOS = new complex <double>**[NUM_TX];

					for (int u = 0; u < NUM_TX; u++)
					{
						channel[bs_idx][ms_idx].CHIR_LOS[u] = new complex <double>*[NUM_RX];

						for (int s = 0; s < NUM_RX; s++)
						{
							channel[bs_idx][ms_idx].CHIR_LOS[u][s] = new complex <double>[MAX_NUM_CLUSTERS];

							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR_LOS[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}
		}
		else
		{
			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					//int adj_sector_num_to_BS = (int)(link[ms_idx].adj_sector[bs_idx] / 3);

					channel[bs_idx][ms_idx].CHIR = new complex <double>**[NUM_TX];

					for (int u = 0; u < NUM_TX; u++)
					{
						channel[bs_idx][ms_idx].CHIR[u] = new complex <double>*[NUM_RX];

						for (int s = 0; s < NUM_RX; s++)
						{
							channel[bs_idx][ms_idx].CHIR[u][s] = new complex <double>[MAX_NUM_CLUSTERS];

							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}

			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					//int adj_sector_num_to_BS = (int)(link[ms_idx].adj_sector[bs_idx] / 3);

					channel[bs_idx][ms_idx].CHIR_LOS = new complex <double>**[NUM_TX];

					for (int u = 0; u < NUM_TX; u++)
					{
						channel[bs_idx][ms_idx].CHIR_LOS[u] = new complex <double>*[NUM_RX];

						for (int s = 0; s < NUM_RX; s++)
						{
							channel[bs_idx][ms_idx].CHIR_LOS[u][s] = new complex <double>[MAX_NUM_CLUSTERS];

							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR_LOS[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}
		}
	}

	else  //// initialize
	{
		//////////////////////// Declare CHIR /////////////////
		///////////////////////////////////////////////////////
		if (TYPE == 11)
		{
			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++) // 
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					for (int u = 0; u < NUM_TX; u++)
					{
						for (int s = 0; s < NUM_RX; s++)
						{
							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}

			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++) // 
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					for (int u = 0; u < NUM_TX; u++)
					{
						for (int s = 0; s < NUM_RX; s++)
						{
							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR_LOS[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}
		}
		else
		{
			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					for (int u = 0; u < NUM_TX; u++)
					{
						for (int s = 0; s < NUM_RX; s++)
						{
							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}

			for (int bs_idx = 0; bs_idx < num_BS; bs_idx++)
			{
				for (int ms_idx = 0; ms_idx < num_MS; ms_idx++)
				{
					for (int u = 0; u < NUM_TX; u++)
					{
						for (int s = 0; s < NUM_RX; s++)
						{
							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)
							{
								channel[bs_idx][ms_idx].CHIR_LOS[u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}
		}
	}



	


	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/*
	if (CH_CAL != 1)   // 1 = run-time calc ϶ ⼭ 
	{
		for (int idx = 0; idx < num_MS; idx++)
		{
			ms[idx].Declare_ch_matrix(); //H_m, HI_m ̸ ϰ ʱȭ, Precalc ⼭ &ʱȭ(ӵؼ), Runtime Ź ʱȭ ʿϹǷ ڿ 
		}
	}
	*/
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}






void Precalculation()
{

	cout << "======== START Declare ============" << endl;

	Declare_Precalc_Parameter();

	cout << "======== END Declare  =============" << endl;


	if (CH_CAL == 0 || CH_CAL == 2) /// 0 = ̸ ϰ run, 2 = ̸길 ϰ 
	{
		Do_precalculation();
	}
	else if (CH_CAL == 3)  ///// 3 = ܺ ҷ 
	{
		Load_precalculation_file();
	}
	





}



void Declare_Precalc_Parameter()
{

	

	//for (int time = 0; time < precalculate_time; time++)        // ð 20ʸŭ
	//{


	/// PreCalc CHIR 
		for (int los = 0; los < 2; los++)            // los 
		{

			for (int indoor = 0; indoor < 2; indoor++)   /////// indoor 
			{
				CHIR_precalc[los][indoor] = new complex <double>***[anglev_mapping_value];

				for (int v = 0; v < anglev_mapping_value; v++)
				{
					CHIR_precalc[los][indoor][v] = new complex <double>**[NUM_TX];

					for (int u = 0; u < NUM_TX; u++)         // TX
					{
						CHIR_precalc[los][indoor][v][u] = new complex <double>*[NUM_RX];

						for (int s = 0; s < NUM_RX; s++) ////// RX 
						{
							CHIR_precalc[los][indoor][v][u][s] = new complex <double>[MAX_NUM_CLUSTERS];

							for (int i = 0; i < MAX_NUM_CLUSTERS; i++)  // cluster 
							{
								CHIR_precalc[los][indoor][v][u][s][i] = complex <double>(0, 0);
							}
						}
					}
				}
			}
		}


	//}
	

	/// Hm precalc 
		/*
	MatrixXcd temp(NUM_RX, NUM_TX);

	for (int r_idx = 0; r_idx < NUM_RX; r_idx++)
	{
		for (int t_idx = 0; t_idx < NUM_TX; t_idx++)
		{
			temp(r_idx, t_idx) = 0.;
		}
	}
	*/
	//for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	//{



		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{

				Hm[los_idx][indoor_idx] = new  MatrixXcd*[anglev_mapping_value];

				for (int v = 0; v < anglev_mapping_value; v++)
				{
					Hm[los_idx][indoor_idx][v] = new MatrixXcd[num_rbs_in_1ms];

					for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
					{
						//Hm[los_idx][indoor_idx][v][freq_idx] = temp;

						Hm[los_idx][indoor_idx][v][freq_idx] = MatrixXcd::Zero(NUM_RX,NUM_TX);
					}
				}
			}
		}



	//}
	

	//// PMI precalc 

	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{

				PMI_precalc[time_idx][los_idx][indoor_idx] = new  int*[anglev_mapping_value];

				for (int v = 0; v < anglev_mapping_value; v++)
				{
					PMI_precalc[time_idx][los_idx][indoor_idx][v] = new int[num_rbs_in_1ms];

					for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
					{

						PMI_precalc[time_idx][los_idx][indoor_idx][v][freq_idx] = 0;

						//PMI_precalc[time_idx][los_idx][indoor_idx][v][freq_idx] = new int[cqi_history_length];

						//for (int cqi_idx = 0; cqi_idx < cqi_history_length; cqi_idx++)
						//{
							//PMI_precalc[time_idx][los_idx][indoor_idx][v][freq_idx][cqi_idx] = 0;
						//}
					}
				}
			}
		}

		cout << "Declare Precalc PMI Time = " << time_idx << endl;
	}
	



	///////////////////////// For sinr estimate precalc 

	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{

				sinr_estimate_precalc[time_idx][los_idx][indoor_idx] = new  double *[anglev_mapping_value];

				for (int v = 0; v < anglev_mapping_value; v++)
				{
					sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v] = new double[num_rbs_in_1ms];

					for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
					{

						sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v][freq_idx] = 0.;


						//sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v][freq_idx] = new double[cqi_history_length];

						//for (int cqi_idx = 0; cqi_idx < cqi_history_length; cqi_idx++)
						//{
							//sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v][freq_idx][cqi_idx] = 0.;
						//}
					}
				}
			}
		}

		cout << "Declare Precalc SINR Time = " << time_idx << endl;
	}
	///////////////////////
	////////////////////////////////// Sum of ch coeff 

	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{

				sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx] = new  double *[anglev_mapping_value];

				for (int v = 0; v < anglev_mapping_value; v++)
				{
					sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v] = new double[num_rbs_in_1ms];

					for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
					{
						sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v][freq_idx] = 0.;
					}

				}
			}
		}

		cout << "Declare CH SUM Time = " << time_idx << endl;
	}


	////////////////////////////////////

}



void Do_precalculation()
{

	cout << "======== START CHIR Precalculation ============" << endl;

	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)   // ޸ Ǵ°  CHIR FT ð   ʵ Ѵ. PMI ð 
	{

		cout << "Precalc Time = " << time_idx << endl;


		channel[0][0].Precalculate(time_idx);  // CHIR  , FT ̿, ///  Ȯο H_m ʿѰ ؾҼ or H_m 
				
		Do_FT_precalc(); // 

		Do_PMI_precalc(time_idx);

		Do_SINR_estimate_precalc(time_idx);

		Do_CH_COEF_SUM_precalc(time_idx);

		//////////// Hm  ////////////////////////
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		///////////  Hm dat   쿡 ϰ ϱ ٶ ///////////////// 

		ofstream Hm_Precalc("Precalc_Hm.dat", ios::app);
		Hm_Precalc << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "row_idx" << "\t" << "col_idx" << "\t" << "Hm_Precalc" << endl;

		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{
						for (int row_idx = 0; row_idx < NUM_RX; row_idx++)
						{
							for (int col_idx = 0; col_idx < NUM_TX; col_idx++)
							{
								Hm_Precalc << time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << row_idx << "\t" << col_idx << "\t" << Hm[los_idx][indoor_idx][v][rbs_idx](row_idx, col_idx) << endl;
							}
						}
						
					}
				}
			}
		}
		
		Hm_Precalc.close();
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////





		///// Hm ʱȭ /////  FT ð Hm  ʱȭ ų ʿ䰡 ִ.
		
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
					{
						Hm[los_idx][indoor_idx][v][freq_idx] = MatrixXcd::Zero(NUM_RX, NUM_TX);
					}
				}
			}
		}
		

	}

	if (CH_CAL == 2)
	{
		cout << "======== START make DAT file ============" << endl;

		Print_DAT_FILE();

		cout << "======== END make DAT file ============" << endl;
	}


	cout << "======== END Precalculation ============" << endl;
	

	

}

void Load_precalculation_file()
{
	cout << "======== START LOADING Precalculation FILE============" << endl;

	//Load_CHIR_precalc_file();

	//Do_FT_precalc();   // FT ׳ ,  ׳ Ȯο, // receiveʿ compute_tone_SINR, MMSE_receiver H_m ϴµ   Լ , ϰԵȴٸ FT ϰų CHIR ޾Ƽ FTϵ (H_m matrix  ָ)

	Load_PMI_precalc_file();

	Load_SINR_estimate_precalc_file();

	Load_CH_COEF_SUM_precalc_file();


	cout << "======== END LOADING Precalculation FILE============" << endl;

}

void Print_DAT_FILE()
{
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	ofstream PMI_Precalc("Precalc_PMI.dat");
	PMI_Precalc << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "PMI_Precalc" << endl;

	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{
						PMI_Precalc << time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << PMI_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;
					}
				}
			}
		}
	

	//cout << "PMI time = " << time_idx + 1 << "ms" << endl;

	}

	PMI_Precalc.close();

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ofstream SINR_ESTIMATE("Precalc_SINR_ESTIMATE.dat");
	SINR_ESTIMATE << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "SINR_ESTIMATE_Precalc" << endl;


	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{
						SINR_ESTIMATE << time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;
					}
				}
			}
		}
	
	}


	SINR_ESTIMATE.close();

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	ofstream CH_COEF_SUM("Precalc_CH_COEF_SUM.dat");
	CH_COEF_SUM << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "CH_COEF_SUM_Precalc" << endl;

	for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{

						CH_COEF_SUM << time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;


					}
				}
			}
		}
	
	}
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	CH_COEF_SUM.close();



}


void Do_FT_precalc()
{


	cout << "START FT" << endl;

	//ofstream FT_Precalc("Precalc_FT.dat");
	//FT_Precalc << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "FT_Precalc" << endl;


	//for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	//{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int r_idx = 0; r_idx < NUM_RX; r_idx++) // Received    Channel
					{
						for (int t_idx = 0; t_idx < NUM_TX; t_idx++) // Transmitted Channel
						{
							int pointer = (int(DFT_SIZE) - (num_rbs_in_1ms*num_freq_per_rbs)) / 2;
							double sum_for_normal = 0;

							for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++) // Subcarrier idx
							{
								int freq_idx = (pointer + (num_freq_per_rbs / 2) + (rbs_idx*num_freq_per_rbs)) * 15000;

								for (int path_idx = 0; path_idx < MAX_NUM_CLUSTERS; path_idx++)
								{
									/// num_path ׳ MAX num cluster  // ̷ ϸ idx  num path Ŀ delay Ⱚ   ־ delay 0 ʱȭŰ 
									/// delay LOS / NLOS ؼ ϵ ϰ channel[0][0].delay_LOS[path_idx] ̷

									complex <double> dft_exp = (0, 0);

									////cout << "channel[0][0].delay[path_idx] = " << channel[0][0].delay[path_idx] << endl;

									if (los_idx == 0)
									{
										// NLOS
										dft_exp = (0.0, -2 * pi*freq_idx*(channel[0][0].delay[path_idx]));

									}
									else
									{
										//LOS
										dft_exp = (0.0, -2 * pi*freq_idx*(channel[0][0].delay_LOS[path_idx]));

									}


									Hm[los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx) += CHIR_precalc[los_idx][indoor_idx][v][t_idx][r_idx][path_idx] * exp(dft_exp);

									////cout << time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx  << endl;
									////cout << Hm[time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;
									////getchar();

								}

								sum_for_normal += pow(abs(Hm[los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx)), 2.);
							}

							sum_for_normal = sum_for_normal / (double)num_rbs_in_1ms;

							for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++) // Subcarrier idx
							{
								Hm[los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx) = Hm[los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx) / sqrt(sum_for_normal);

								//FT_Precalc << time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << Hm[time_idx][los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx) << endl;

							}
						}
					}
				}


			}
		}

		//cout << "FT Time = " << time_idx + 1 << "ms" << endl;
	//}

	//FT_Precalc.close();

}

void Do_PMI_precalc(int _time_idx)
{

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////PMI  κ //////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	cout << "START PMI" << endl;


	//ofstream PMI_Precalc("Precalc_PMI.dat");
	//PMI_Precalc << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t"  << "PMI_Precalc" << endl;



	//for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	//{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{

						if (NUM_TX != 4)    // TX 4 ƴϸ PMI  ϵ.. (For Calibration 3GPP TR 36.814)
						{
							PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] = 0;   // ׳ 0   
						}
						else if (true) //////////////////////  SVD ̿Ͽ PMI  ( ڵ ) /////////////////////////////////////////////////////
						{
							JacobiSVD <MatrixXcd> svd(Hm[los_idx][indoor_idx][v][rbs_idx], ComputeFullU | ComputeFullV);    ////// U = 2x2, V = 4x4

							/*
							cout << "Its singular values are:" << endl << svd.singularValues() << endl;
							cout << "Its left singular vectors are the columns of the thin U matrix:" << endl << svd.matrixU() << endl;
							cout << "Its right singular vectors are the columns of the thin V matrix:" << endl << svd.matrixV() << endl;

							getchar();
							*/

							MatrixXcd H_eq;

							H_eq = ( svd.matrixV().col(0) ).adjoint();  ///////// 1x4 vector

							//cout << "H_eq = " << H_eq << endl;

							double max = -500;

							for (int cw_idx = 0; cw_idx < 64; cw_idx++)
							{
								double temp = (H_eq * code_word[cw_idx].conjugate()).norm();

								if (temp > max)
								{
									max = temp;
									PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] = cw_idx;
								}
							}

							//cout << "Precalc_PMI = " << PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;
							//getchar();


						}
						else  ///////// PMI ϴ ٸ  /////////////////////
						{
							Matrix4cd Ohm;

							double min = 500;

							for (int cw_idx = 0; cw_idx < 16; cw_idx++)
							{

								Ohm = WW[cw_idx] * (Hm[los_idx][indoor_idx][v][rbs_idx].adjoint() * Hm[los_idx][indoor_idx][v][rbs_idx]) * WW[cw_idx].adjoint(); //Ohm = W HH H WH // 4x4 matrix

								Matrix4cd UppOhm;
								Matrix4cd LowOhm;

								UppOhm = Ohm.triangularView<StrictlyUpper>(); //diagonal   triangle matrix 
								LowOhm = Ohm.triangularView<StrictlyLower>(); //diagonal  Ʒ triangle matrix 



								double temp = UppOhm.norm() + LowOhm.norm(); // minimizes off diagonal entries of Ohm

								if (temp < min)
								{
									min = temp;
									//PMI_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx][time_idx % cqi_history_length] = cw_idx;
									PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] = cw_idx;
								}


								//PMI_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx][time_idx % cqi_history_length] = cw_idx;
							}
						}


					

					//	PMI_Precalc << _time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << _time_idx % cqi_history_length << "\t" << PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;

					}
				}
			}
		}

		//cout << "PMI time = " << time_idx + 1 << "ms" << endl;

	//}

	//PMI_Precalc.close();

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


}

void Do_SINR_estimate_precalc(int _time_idx)
{
	//////////////////////////////// SINR estimate /////////////////////////////////////////////////////////////////////////////////////////////////////

	cout << "START SINR estimate" << endl;

//	ofstream SINR_ESTIMATE("Precalc_SINR_ESTIMATE.dat");
	//SINR_ESTIMATE << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "SINR_ESTIMATE_Precalc" << endl;

	//for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	//{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{

						if (NUM_TX != 4) /// For Calibration 1x2 SIMO
						{
							sinr_estimate_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] = 0.;   ///  1x2 cali Ⱦ(CH COEFF SUM ϹǷ) ׳ 0 
						}
						else
						{
							MatrixXcd A;
							MatrixXcd B;
							VectorXcd w;
							Matrix2cd eye = Matrix2cd::Identity();

							//w = code_word[PMI_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx][time_idx % cqi_history_length]].conjugate();   // 4x1 vector
							w = code_word[PMI_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx]].conjugate();   // 4x1 vector
							A = Hm[los_idx][indoor_idx][v][rbs_idx] * w; // * UE PATHLOSS  // 2x4 4x1
							B = eye;   /// 2x2 indentity  // * UE ICI

							/*
							cout << "B norm = " << B.norm() << endl;
							getchar();
							getchar();
							*/
							/// eigen matrix norm ܼ  ׸ ؼ ؼ Ʈ, itpp 2-norm ٸ Ǵ°Ͱ. 
							// itpp B.norm()   Ȯغʿ䰡 ִٸ B identity̹Ƿ 2-norm 1ϰ -> 1 

							// A norm  norm̶ ؼ ؼ Ʈ ´. ״ ξ 

							//sinr_estimate_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] = A.norm() / B.norm();
							sinr_estimate_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] = A.norm();


							//SINR_ESTIMATE << _time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << _time_idx % cqi_history_length << "\t" << sinr_estimate_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;


						}
					}
				}
			}
		}
	//}


	//SINR_ESTIMATE.close();

}

void Do_CH_COEF_SUM_precalc(int _time_idx)
{
	//////////////////////////////// FOR Received SINR -> sum of ch coefficeint ////////////////////////////////////////////////
	cout << "START SUM of CH COEFF" << endl;

	//ofstream CH_COEF_SUM("Precalc_CH_COEF_SUM.dat");
	//CH_COEF_SUM << "time_idx" << "\t" << "LOS" << "\t" << "INDOOR" << "\t" << "V_ANGLE_MAPPING" << "\t" << "RB_idx" << "\t" << "CH_COEF_SUM_Precalc" << endl;

	//for (int time_idx = 0; time_idx < precalculate_time; time_idx++)
	//{
		for (int los_idx = 0; los_idx < 2; los_idx++)
		{
			for (int indoor_idx = 0; indoor_idx < 2; indoor_idx++)
			{
				for (int v = 0; v < anglev_mapping_value; v++)
				{
					for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
					{
						for (int r_idx = 0; r_idx < NUM_RX; r_idx++) // Received    Channel
						{
							for (int t_idx = 0; t_idx < NUM_TX; t_idx++) // Transmitted Channel
							{
								sum_of_ch_coef_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] += pow(abs(Hm[los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx)), 2.);

								/*
								cout << "pow(abs(Hm[time_idx][los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx)), 2.) ==" << pow(abs(Hm[time_idx][los_idx][indoor_idx][v][rbs_idx](r_idx, t_idx)), 2.) << endl;
								cout << "sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx] == " << sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;
								getchar();
								*/
							}
						}

						//CH_COEF_SUM << _time_idx << "\t" << los_idx << "\t" << indoor_idx << "\t" << v << "\t" << rbs_idx << "\t" << sum_of_ch_coef_precalc[_time_idx][los_idx][indoor_idx][v][rbs_idx] << endl;


					}
				}
			}
		}
	//}
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//CH_COEF_SUM.close();

}


void Load_CHIR_precalc_file()
{

	/*
	string trash;

	int time_idx = 0;
	int los_idx = 0;
	int indoor_idx = 0;
	int v_idx = 0;
	int TX_idx = 0;
	int RX_idx = 0;
	int cluster_idx = 0;
	complex <double> CHIR_value = (0, 0);
	int loop = 0;

	ifstream Precalc_CHIR_input;
	Precalc_CHIR_input.open("Precalc_CHIR.DAT");

	if (Precalc_CHIR_input.is_open())
	{
		cout << "=== Precalc_CHIR.DAT file open ===" << endl;
	}
	else
	{
		cout << "=== NO FILE!!!!!! Please CHECK ===" << endl;
		getchar();
	}

	while (!Precalc_CHIR_input.eof())
	{
		if (loop == 0)  // ù ׳ ѱ
		{
			for (int i = 0; i < 8; i++)
			{
				Precalc_CHIR_input >> trash;
			}

			loop++;
		}
		else
		{
			Precalc_CHIR_input >> time_idx;
			Precalc_CHIR_input >> los_idx;
			Precalc_CHIR_input >> indoor_idx;
			Precalc_CHIR_input >> v_idx;
			Precalc_CHIR_input >> TX_idx;
			Precalc_CHIR_input >> RX_idx;
			Precalc_CHIR_input >> cluster_idx;
			Precalc_CHIR_input >> CHIR_value;

			CHIR_precalc[time_idx][los_idx][indoor_idx][v_idx][TX_idx][RX_idx][cluster_idx] = CHIR_value;



			loop++;
		}
	}

	cout << "=== Precalc_CHIR.DAT load Ϸ ===" << endl;
	Precalc_CHIR_input.close();

	*/

}

void Load_PMI_precalc_file()
{
	string trash;

	int time_idx = 0;
	int los_idx = 0;
	int indoor_idx = 0;
	int v_idx = 0;
	int RB_idx = 0;
	int PMI_value = 0;
	int loop = 0;

	ifstream Precalc_PMI_input;
	Precalc_PMI_input.open("Precalc_PMI.DAT");

	if (Precalc_PMI_input.is_open())
	{
		cout << "=== Precalc_PMI.DAT file open ===" << endl;
	}
	else
	{
		cout << "=== NO FILE!!!!!! Please CHECK ===" << endl;
		getchar();
	}

	while (!Precalc_PMI_input.eof())
	{
		if (loop == 0)  // ù ׳ ѱ
		{
			for (int i = 0; i < 6; i++)
			{
				Precalc_PMI_input >> trash;
			}

			loop++;
		}
		else
		{
			Precalc_PMI_input >> time_idx;
			Precalc_PMI_input >> los_idx;
			Precalc_PMI_input >> indoor_idx;
			Precalc_PMI_input >> v_idx;
			Precalc_PMI_input >> RB_idx;
			Precalc_PMI_input >> PMI_value;



			//PMI_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx][CQI_length_idx] = PMI_value;
			PMI_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx] = PMI_value;

			/*
			if (time_idx > 1740)
			{
				cout << "time = " << time_idx << endl;
				cout << "los = " << los_idx << endl;
				cout << "indoor_idx = " << indoor_idx << endl;
				cout << "v_idx = " << v_idx << endl;
				cout << "PMI_value = " << PMI_value << endl;
				cout << "PMI_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx] = " << PMI_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx] << endl;

			}
			*/
	

			//getchar();
			

			loop++;
		}
	}

	cout << "=== Precalc_PMI.DAT load Ϸ ===" << endl;
	Precalc_PMI_input.close();

}

void Load_SINR_estimate_precalc_file()
{
	string trash;

	int time_idx = 0;
	int los_idx = 0;
	int indoor_idx = 0;
	int v_idx = 0;
	int RB_idx = 0;
	double SINR_value = 0.;
	int loop = 0;

	ifstream Precalc_SINR_input;
	Precalc_SINR_input.open("Precalc_SINR_ESTIMATE.DAT");

	if (Precalc_SINR_input.is_open())
	{
		cout << "=== Precalc_SINR_ESTIMATE.DAT file open ===" << endl;
	}
	else
	{
		cout << "=== NO FILE!!!!!! Please CHECK ===" << endl;
		getchar();
	}

	while (!Precalc_SINR_input.eof())
	{
		if (loop == 0)  // ù ׳ ѱ
		{
			for (int i = 0; i < 6; i++)
			{
				Precalc_SINR_input >> trash;
			}

			loop++;
		}
		else
		{
			Precalc_SINR_input >> time_idx;
			Precalc_SINR_input >> los_idx;
			Precalc_SINR_input >> indoor_idx;
			Precalc_SINR_input >> v_idx;
			Precalc_SINR_input >> RB_idx;
			Precalc_SINR_input >> SINR_value;


			//sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx][CQI_length_idx] = SINR_value;
			sinr_estimate_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx] = SINR_value;
			/*
			cout << "time = " << time_idx << endl;
			cout << "los = " << los_idx << endl;
			cout << "indoor_idx = " << indoor_idx << endl;
			cout << "v_idx = " << v_idx << endl;

			getchar();
			*/

			loop++;
		}
	}

	cout << "=== Precalc_SINR_ESTIMATE.DAT load Ϸ ===" << endl;
	Precalc_SINR_input.close();

}

void Load_CH_COEF_SUM_precalc_file()
{
	string trash;

	int time_idx = 0;
	int los_idx = 0;
	int indoor_idx = 0;
	int v_idx = 0;
	int RB_idx = 0;

	double CH_SUM_value = 0.;
	int loop = 0;

	ifstream Precalc_CH_SUM_input;
	Precalc_CH_SUM_input.open("Precalc_CH_COEF_SUM.DAT");

	if (Precalc_CH_SUM_input.is_open())
	{
		cout << "=== Precalc_CH_COEF_SUM.DAT file open ===" << endl;
	}
	else
	{
		cout << "=== NO FILE!!!!!! Please CHECK ===" << endl;
		getchar();
	}

	while (!Precalc_CH_SUM_input.eof())
	{
		if (loop == 0)  // ù ׳ ѱ
		{
			for (int i = 0; i < 6; i++)
			{
				Precalc_CH_SUM_input >> trash;
			}

			loop++;
		}
		else
		{
			Precalc_CH_SUM_input >> time_idx;
			Precalc_CH_SUM_input >> los_idx;
			Precalc_CH_SUM_input >> indoor_idx;
			Precalc_CH_SUM_input >> v_idx;
			Precalc_CH_SUM_input >> RB_idx;
			Precalc_CH_SUM_input >> CH_SUM_value;


			sum_of_ch_coef_precalc[time_idx][los_idx][indoor_idx][v_idx][RB_idx] = CH_SUM_value;

			/*
			cout << "time = " << time_idx << endl;
			cout << "los = " << los_idx << endl;
			cout << "indoor_idx = " << indoor_idx << endl;
			cout << "v_idx = " << v_idx << endl;

			getchar();
			*/

			loop++;
		}
	}

	cout << "=== Precalc_CH_COEF_SUM.DAT load Ϸ ===" << endl;
	Precalc_CH_SUM_input.close();

}


